/*

TextInterface.java 
  -- classes implementing a text based interface to iasolver,
     as part of the "ia_solver library" version 0.1beta1, 10/97

Copyright (C) 1997 Timothy J. Hickey

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

 */


package ia_solver;

import java.awt.*;
import java.io.DataInputStream;
import java.lang.*;
import java.util.*;

import ia_parser.*;
import ia_math.RealInterval;
/* import ia_solver.utility.*; */

/**
 * this implements a text interface to the IAsolver.
 */

class TextInterface implements SolverGUI {
  Exp exp=null; 
  int max_narrow = 100;
  String intervalStyle="x +/- e";
  RealIntervalTable vars= new RealIntervalTable();
  Vector var_stack = new Vector();
  boolean success= true;

  /**
   * this prompts the user and reads a constraint set.
   * The constraint set is terminated by an empty line.
   */
  private static String readConstraints(DataInputStream input) {
    String constraints="", tmpString;
    try {
      System.out.print("|  "); System.out.flush();
      constraints = "";
      while ((tmpString = input.readLine()).length()>0) {
        constraints = constraints + tmpString + " ";
        System.out.print("|  "); System.out.flush();
      }
    } catch (Exception e) {
      System.out.println("Parsing Error: "+e+"\n"+
         " Please enter a constraint\n");
    }
    return constraints;
  }

  private boolean absolve_hi(String xx,double amt) {
    double newhi, newlo, oldhi, oldlo;
    RealInterval x;

    //    System.out.println("\n\nTrying to absolve upper bound of "
    //      +this.vars.getNthVarName(i));

    /* first we push a copy of the current environment */
    this.var_stack.addElement(this.vars.clone());
    //    System.out.println("stack = "+this.var_stack);

    /* next we try to absolve by 10% on the top */
    x = this.vars.lookup(xx);
    oldlo = x.lo(); oldhi = x.hi();
    newhi = ((1.0-amt)*oldhi + amt* oldlo);
    newlo = (amt*oldhi + (1.0-amt)* oldlo);
    x.intersect(new RealInterval(newhi,oldhi));
    //    System.out.println("tmp interval is ["+newhi+","+oldhi+"]");

    Interpreter interpreter = new Interpreter(Interpreter.NARROW,this);
    success=true;
    interpreter.setPriority(Thread.NORM_PRIORITY+1);
    this.exp.bindVars(this.vars);
    interpreter.narrow();

    int s = var_stack.size();
    this.vars = (RealIntervalTable) this.var_stack.lastElement();
    this.var_stack.removeElementAt(s-1);

    if (success == false) {
            System.out.println(
               "\nNarrowing upper bound of "+ xx +
               " from " + oldhi + " to "+newhi);
      this.vars.lookup(xx).intersect(new RealInterval(oldlo,newhi));
    }
    else {
      System.out.println("\n     Kept upper bound of "+ xx);
      return false;
    }

    success = true;
    this.exp.bindVars(this.vars);

    System.out.println(
        "current bindings:\n "+this.vars +"\n");

    return true;

  }

  private boolean absolve_lo(String xx,double amt) {
    double newhi, newlo, oldhi, oldlo;
    RealInterval x;

        System.out.println("\n\nTrying to absolve lower bound of "
          +xx);

    /* first we push a copy of the current environment */
    this.var_stack.addElement(this.vars.clone());
        System.out.println("stack = "+this.var_stack);

    /* next we try to absolve by 10% on the top */
    x = this.vars.lookup(xx);
    oldlo = x.lo(); oldhi = x.hi();
    newhi = ((1-amt)*oldhi + amt* oldlo);
    newlo = (amt*oldhi + (1.0-amt)* oldlo);
    x.intersect(new RealInterval(oldlo,newlo));
        System.out.println("tmp interval is ["+oldlo+","+newlo+"]");

    Interpreter interpreter = new Interpreter(Interpreter.NARROW,this);
    success=true;
    interpreter.setPriority(Thread.NORM_PRIORITY+1);
    this.exp.bindVars(this.vars);
    interpreter.narrow();

    int s = var_stack.size();
    this.vars = (RealIntervalTable) this.var_stack.lastElement();
    this.var_stack.removeElementAt(s-1);

    if (success == false) {
            System.out.println(
               "\nNarrowing lower bound of "+
               xx + " from " + oldlo + " to "+newlo);
      this.vars.lookup(xx).intersect(new RealInterval(newlo,oldhi));
    }
    else {
      System.out.println("\n     Kept lower bound of "+ xx);
      return false;
    }
    
    success = true;
    this.exp.bindVars(this.vars);

    System.out.println(
        "current bindings:\n "+this.vars +"\n");

    return true;

  }

  public void absolve() {
    String[] varlist =  new String[vars.size()];

    for(int i=0;i<vars.size();i++) {
       varlist[i] = this.vars.getNthVarName(i);
    }

    for(int i=0;i<vars.size();i++) {
      absolve_hi(varlist[i],0.5);
      absolve_lo(varlist[i],0.5);
    }
  }


  /**
   * A simple read/narrow/print loop.
   */
  public static void interpret() {
    DataInputStream input = new DataInputStream(System.in);
    String constraints="",tmpString;
    Interpreter interpreter;
    TextInterface ti = new TextInterface();

    System.out.println(
    "\n\n\n     Brandeis IAsolver (version 0.1beta1,10/97)\n\n\n");

    while (true) {
      int i;

      System.out.print("\n>> "); System.out.flush();

      try {tmpString=input.readLine();}
      catch(Exception e) {
        System.out.println("Syntax error: try again");continue;
      }

      if (tmpString.equals("constrain")) {
          constraints = constraints + readConstraints(input);
          if (constraints == "") continue;
          System.out.println(" Constraint set = \n"+constraints+"\n\n");
          ti.exp = Parser.parseString(constraints);
          if (ti.exp==null) {
            System.out.println("Syntax error in constraints\n\n>>");
            continue;
	  }else System.out.println("Just parsed \n"+ti.exp+"\n\n");
          ti.exp.bindVars(ti.vars);
      }else if (tmpString.equals("narrow")) {
          interpreter = new Interpreter(ti);
          ti.success=true;
          interpreter.setPriority(Thread.NORM_PRIORITY+1);
          ti.exp.bindVars(ti.vars);
          interpreter.narrow();
          if (ti.success)
            System.out.println("Narrowing completed");
          else
            System.out.println("Constraints are unsatisfiable");
      }else if (tmpString.equals("absolve")) {
          ti.absolve();
      } else if (tmpString.equals("print")) {
          System.out.println(ti.vars);
      } else if (tmpString.equals("reset")) {
          constraints=""; ti.exp=null; ti.vars = new RealIntervalTable();
      } else if (tmpString.equals("push")) {
          ti.var_stack.addElement(ti.vars.clone());
          ti.var_stack.addElement(constraints);
          System.out.println("stack = "+ti.var_stack);
      } else if (tmpString.equals("pop")) {
          int s = ti.var_stack.size();
          if (s==0) {
	    System.out.println("can't pop an empty stack");
            continue;
	  }
          constraints = (String) ti.var_stack.lastElement();
          ti.exp = Parser.parseString(constraints);
          ti.var_stack.removeElementAt(s-1);

          ti.vars = (RealIntervalTable) ti.var_stack.lastElement();
          ti.var_stack.removeElementAt(s-2);

          if (ti.exp != null)
            ti.exp.bindVars(ti.vars);
      } else if (tmpString.equals("stack")) {
          System.out.println("stack = \n"+ti.var_stack);
      } else if (tmpString.equals("constraints")) {
          System.out.println("constraints = \n"+constraints);
      } else if (tmpString.equals("quit")) 
         break;
      else if (tmpString.equals("exit")) 
         break;
      else if (tmpString.equals("help")) 
         System.out.println(
           "Commands: constrain narrow print\n"+
           "          push pop stack constraints\n"+
           "          quit exit bye help\n");
      else if (tmpString.equals("bye")) 
         break;
      else {
        System.out.println("unknown command!"); continue;
      }


    }
    System.out.println("\n\nByebye\n\n\n");
  }

  public void setMessageField(String s){
    //    System.out.println("Message: "+s);
  }

  public void setButtonStatus(String s){
    //    System.out.println("Buttons: "+s);
  }

  public Integer getMaxNarrow(){
    return new Integer(max_narrow);
  }

  public Exp getConstraints(){
    return exp;
  }

  public RealIntervalTable getVars() {
    return vars;
  }

  public String getIntervalStyle(){
      return intervalStyle;
  }

  public void setVarList(String s){
    //    System.out.println("Variables: \n"+s+"\n");
  }

  public void notifyNarrows(int i) {
    if (i%5==0) {
      System.out.print(".");
      System.out.flush();
    }
    //    if (i%99==0) System.out.println(this.vars);
  }


  public void notifyFailure(){
    vars = new RealIntervalTable();
    success = false;
  }

  public void notifySuccess(){
    success = true;
  }

  public void pushStack(){
  }
  public boolean popStack(){
    return(false);
  }
  public int sizeStack(){
    return(0);
  }

  public static void main(String[] args) {
    interpret();
  }


}
